'''
Created on Sep 24, 2013
Updated on Mar 26, 2017

@author: paepcke

'''

# To run these tests, you need a local MySQL server,
# an account called 'unittest' on that server, and a
# database called 'unittest', also on that server.
#
# User unittest must be able to log in without a password, 
# the user only needs permissions for the unittest database.
# Here are the necessary MySQL commands:
#
#   CREATE DATABASE unittest;
#   CREATE USER unittest@localhost;
#   GRANT 'SELECT', 'INSERT', 'UPDATE', 
#         'DELETE', 'CREATE', 'CREATE TEMPORARY TABLES', 
#         'DROP', 'ALTER' ON unittest.* TO unittest@localhost;
#
# The tests are designed to work on MySQL 5.6 and 5.7.

TEST_ALL = True
#TEST_ALL = False


from collections import OrderedDict
import re
import socket
import subprocess
import unittest

from shutilwhich import which

from pymysql_utils import MySQLDB, DupKeyAction, no_warn_no_table
import pymysql_utils


class TestPymysqlUtils(unittest.TestCase):
    '''
    Tests pymysql_utils.    
    '''
  
    @classmethod
    def setUpClass(cls):
        # Ensure that a user unittest with the proper
        # permissions exists in the db:
        TestPymysqlUtils.env_ok = True
        TestPymysqlUtils.err_msg = ''
        try:
            needed_grants = ['SELECT', 'INSERT', 'UPDATE', 
                             'DELETE', 'CREATE', 'CREATE TEMPORARY TABLES', 
                             'DROP', 'ALTER']
            mysqldb = MySQLDB(host='localhost', port=3306, user='unittest', db='unittest')
            grant_query = 'SHOW GRANTS FOR unittest@localhost'
            query_it = mysqldb.query(grant_query)
            # First row of the SHOW GRANTS response should be:
            first_grant = "GRANT USAGE ON *.* TO 'unittest'@'localhost'"
            # Second row depends on the order in which the 
            # grants were provided. The row will look something
            # like:
            #   GRANT SELECT, INSERT, UPDATE, DELETE, ..., CREATE, DROP, ALTER ON `unittest`.* TO 'unittest'@'localhost'
            # Verify:
            if query_it.next() != first_grant:
                TestPymysqlUtils.err_msg = '''
                    User 'unittest' is missing USAGE grant needed to run the tests.
                    Also need this in your MySQL: 
                    
                          %s
                    ''' % 'GRANT %s ON unittest.* TO unittest@localhost' % ','.join(needed_grants)
                TestPymysqlUtils.env_ok = False
                return
            grants_str = query_it.next()
            for needed_grant in needed_grants:
                if grants_str.find(needed_grant) == -1:
                    TestPymysqlUtils.err_msg = '''
                    User 'unittest' does not have the '%s' permission needed to run the tests.
                    Need this in your MySQL:
                    
                        %s
                    ''' % (needed_grant, 'GRANT %s ON unittest.* TO unittest@localhost;' % ','.join(needed_grants))
                    TestPymysqlUtils.env_ok = False
                    return  
        except ValueError:
            TestPymysqlUtils.err_msg = '''
               For unit testing, localhost MySQL server must have 
               user 'unittest' without password, and a database 
               called 'unittest'. To create these prerequisites 
               in MySQL:
               
                    CREATE USER unittest@localhost;
                    CREATE DATABASE unittest; 
               This user needs permissions:
                    %s 
               ''' % 'GRANT %s ON unittest.* TO unittest@localhost;' % ','.join(needed_grants)
            TestPymysqlUtils.env_ok = False

        # Check MySQL version:
        try:
            (major, minor) = TestPymysqlUtils.get_mysql_version()
        except Exception as e:
            raise OSError('Could not get mysql version number: %s' % str(e))
            
        if major is None:
            print('Warning: MySQL version number not found; testing as if V5.7')
            TestPymysqlUtils.major = 5
            TestPymysqlUtils.minor = 7
        else:
            TestPymysqlUtils.major = major
            TestPymysqlUtils.minor = minor
            if major != 5 or (minor != 6 and minor != 7):
                print('Warning: MySQL version is %s.%s; but testing as if V5.7')
        

    def setUp(self):
        if not TestPymysqlUtils.env_ok:
            raise RuntimeError(TestPymysqlUtils.err_msg)
        try:
            self.mysqldb = MySQLDB(host='localhost', port=3306, user='unittest', db='unittest')
        except ValueError as e:
            self.fail(str(e) + " (For unit testing, localhost MySQL server must have user 'unittest' without password, and a database called 'unittest')")
            
        # Make MySQL version more convenient to check:
        if TestPymysqlUtils.minor >= 7:
            self.mysql_ge_7 = True
        else:
            self.mysql_ge_7 = False


    def tearDown(self):
        self.mysqldb.dropTable('unittest')
        # Make sure the test didn't set a password
        # for user unittest in the db:
        self.mysqldb.execute("SET PASSWORD FOR unittest@localhost = '';")
        self.mysqldb.close()

    # ----------------------- Table Manilupation -------------------------

    #-------------------------
    # Creating and Dropping Tables 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testCreateAndDropTable(self):
        mySchema = {
          'col1' : 'INT',
          'col2' : 'varchar(255)',
          'col3' : 'FLOAT',
          'col4' : 'TEXT',
          #'col5' : 'JSON'  # Only works MySQL 5.7 and up.
          }
        self.mysqldb.createTable('myTbl', mySchema, temporary=False)
        tbl_desc = subprocess.check_output([self.mysqldb.mysql_loc,
                                           '-u',
                                           'unittest',
                                           'unittest',
                                           '-e',
                                           'DESC myTbl;'])
        expected = 'Field\tType\tNull\tKey\tDefault\tExtra\n' + \
                   'col4\ttext\tYES\t\tNULL\t\n' + \
                   'col2\tvarchar(255)\tYES\t\tNULL\t\n' + \
                   'col3\tfloat\tYES\t\tNULL\t\n' + \
                   'col1\tint(11)\tYES\t\tNULL\t\n'
        self.assertEqual(tbl_desc, expected)
        
        # Query mysql information schema to check for table
        # present. Use raw cursor to test independently from
        # the pymysql_utils query() method:
        
        self.mysqldb.dropTable('myTbl')
        cursor = self.mysqldb.connection.cursor()
        tbl_exists_query = '''
                  SELECT table_name 
                    FROM information_schema.tables 
                   WHERE table_schema = 'unittest' 
                     AND table_name = 'myTbl';
                     '''
        cursor.execute(tbl_exists_query)
        self.assertEqual(cursor.rowcount, 0)
        cursor.close()

    #-------------------------
    # Creating Temporary Tables 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testCreateTempTable(self):
        mySchema = {
          'col1' : 'INT',
          'col2' : 'varchar(255)',
          'col3' : 'FLOAT',
          'col4' : 'TEXT',
          #'col5' : 'JSON'  # Only works MySQL 5.7 and up.
          }
        self.mysqldb.createTable('myTbl', mySchema, temporary=True)
        
        # Check that tbl exists.
        # NOTE: can't use query to mysql.informationschema,
        # b/c temp tables aren't listed there.
        
        try:
            # Will return some tuple; we don't
            # care what exaclty, as long as the
            # cmd doesn't fail:
            self.mysqldb.query('DESC myTbl').next()
        except Exception:
            self.fail('Temporary table not found after creation.')
        
        # Start new session, which should remove the table.
        # Query mysql information schema to check for table
        # present. Use raw cursor to test independently from
        # the pymysql_utils query() method:
        
        self.mysqldb.close()

        try:
            self.mysqldb = MySQLDB(host='localhost', port=3306, user='unittest', db='unittest')
        except ValueError as e:
            self.fail(str(e) + "Could not re-establish MySQL connection.")

        # NOTE: can't use query to mysql.informationschema,
        # b/c temp tables aren't listed there.
        
        try:
            self.mysqldb.query('DESC myTbl').next()
            self.fail("Temporary table did not disappear with session exit.")
        except ValueError:
            pass


    #-------------------------
    # Table Truncation 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testTruncate(self):
      
        # Initial test db with known num of rows:
        rows_in_test_db = self.buildSmallDb()
        cursor = self.mysqldb.connection.cursor()
        cursor.execute('SELECT * FROM unittest;')
        self.assertEqual(cursor.rowcount, rows_in_test_db)
        
        self.mysqldb.truncateTable('unittest')
        
        cursor.execute('SELECT * FROM unittest;')
        self.assertEqual(cursor.rowcount, 0)
        cursor.close()

    # ----------------------- Insertion and Update -------------------------
    
    #-------------------------
    # Insert One Row 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testInsert(self):
        schema = OrderedDict([('col1', 'INT'), ('col2', 'TEXT')])
        self.mysqldb.createTable('unittest', schema)
        colnameValueDict = OrderedDict([('col1', 10)])
        self.mysqldb.insert('unittest', colnameValueDict)
        self.assertEqual((10, None), self.mysqldb.query("SELECT * FROM unittest").next())
        # for value in self.mysqldb.query("SELECT * FROM unittest"):
        #    print value
 
    #-------------------------
    # Insert One Row With Error 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testInsertWithError(self):
        schema = OrderedDict([('col1', 'INT'), ('col2', 'TEXT')])
        self.mysqldb.createTable('unittest', schema)
        colnameValueDict = OrderedDict([('col1', 10)])
        self.mysqldb.insert('unittest', colnameValueDict)
        self.assertEqual((10, None), self.mysqldb.query("SELECT * FROM unittest").next())
        # for value in self.mysqldb.query("SELECT * FROM unittest"):
        #    print value

    
    #-------------------------
    # Insert Several Columns 
    #--------------

    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testInsertSeveralColumns(self):
        schema = OrderedDict([('col1', 'INT'), ('col2', 'TEXT')])
        self.mysqldb.createTable('unittest', schema)
        colnameValueDict = OrderedDict([('col1', 10), ('col2', 'My Poem')])
        self.mysqldb.insert('unittest', colnameValueDict)
        res = self.mysqldb.query("SELECT * FROM unittest").next()
        self.assertEqual((10, 'My Poem'), res)
    

    #-------------------------
    # Bulk Insertion 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testBulkInsert(self):
        # Called twice: once by the unittest engine,
        # and again by testWithMySQLPassword() to 
        # exercise the pwd-bound branch in bulkInsert().
      
        # Build test db (this already tests basic bulkinsert):
        #                  col1   col2
        #                   10,  'col1'
        #                   20,  'col2'
        #                   30,  'col3'
        self.buildSmallDb()
        self.mysqldb.execute('ALTER TABLE unittest ADD PRIMARY KEY(col1)')
        
        # Provoke a MySQL error: duplicate primary key (i.e. 10): 
        # Add another row:  10,  'newCol1':
        colNames = ['col1', 'col2']
        colValues = [(10, 'newCol1')]
        
        (errors, warnings) = self.mysqldb.bulkInsert('unittest', colNames, colValues) #@UnusedVariable
        
        # For MySQL 5.7, expect something like:
        #    ((u'Warning', 1062L, u"Duplicate entry '10' for key 'PRIMARY'"),)
        # MySQL 5.6 just skips: 
        
        if self.mysql_ge_7:
            self.assertEqual(len(warnings), 1)
        else:
            self.assertIsNone(warnings)
            
        # First tuple should still be (10, 'col1'):
        self.assertEqual('col1', self.mysqldb.query('SELECT col2 FROM unittest WHERE col1 = 10').next())
        
        # Try update again, but with replacement:
        (errors, warnings) = self.mysqldb.bulkInsert('unittest', colNames, colValues, onDupKey=DupKeyAction.REPLACE) #@UnusedVariable
        self.assertIsNone(warnings)
        # Now row should have changed:
        self.assertEqual('newCol1', self.mysqldb.query('SELECT col2 FROM unittest WHERE col1 = 10').next())
        
        # Insert a row with duplicate key, specifying IGNORE:
        colNames = ['col1', 'col2']
        colValues = [(10, 'newCol2')]
        (errors, warnings) = self.mysqldb.bulkInsert('unittest', colNames, colValues, onDupKey=DupKeyAction.IGNORE) #@UnusedVariable
        # Even when ignoring dup keys, MySQL 5.7 issues a warning
        # for each dup key:
        
        if self.mysql_ge_7:
            self.assertEqual(len(warnings), 1)
        else:
            self.assertIsNone(warnings)
        
        self.assertEqual('newCol1', self.mysqldb.query('SELECT col2 FROM unittest WHERE col1 = 10').next())
        
        # Provoke an error:
        colNames = ['col1', 'col2', 'col3']
        colValues = [(10, 'newCol2')]
        (errors, warnings) = self.mysqldb.bulkInsert('unittest', colNames, colValues, onDupKey=DupKeyAction.IGNORE) #@UnusedVariable
        self.assertEqual(len(errors), 1)
        
    #-------------------------
    # Updates 
    #--------------

    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testUpdate(self):
      
        num_rows = self.buildSmallDb()
        cursor = self.mysqldb.connection.cursor()
        
        # Initially, col2 of row0 must be 'col1':
        cursor.execute('SELECT col2 FROM unittest WHERE col1 = 10')
        col2_row_zero = cursor.fetchone()
        self.assertTupleEqual(col2_row_zero, ('col1',))
        
        self.mysqldb.update('unittest', 'col1', 40, fromCondition='col1 = 10')
        
        # Now no col1 with value 10 should exist:
        cursor.execute('SELECT col2 FROM unittest WHERE col1 = 10')
        self.assertEqual(cursor.rowcount, 0)
        # But a row with col1 == 40 should have col2 == 'col1':
        cursor.execute('SELECT col2 FROM unittest WHERE col1 = 40')
        col2_res = cursor.fetchone()
        self.assertTupleEqual(col2_res, ('col1',))
        
        # Update *all* rows in one column:
        self.mysqldb.update('unittest', 'col1', 0)
        cursor.execute('SELECT count(*) FROM unittest WHERE col1 = 0')
        res_count = cursor.fetchone()
        self.assertTupleEqual(res_count, (num_rows,))

        # Provoke an error:
        (errors,warnings) = self.mysqldb.update('unittest', 'col6', 40, fromCondition='col1 = 10') #@UnusedVariable
        self.assertEqual(len(errors), 1)
    
    # ----------------------- Queries -------------------------         

    #-------------------------
    # Query With Result Iteration 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testQueryIterator(self):
        self.buildSmallDb()

        for rowNum, result in enumerate(self.mysqldb.query('SELECT col1,col2 FROM unittest')):
            if rowNum == 0:
                self.assertEqual((10, 'col1'), result)
            elif rowNum == 1:
                self.assertEqual((20, 'col2'), result)
            elif rowNum == 2:
                self.assertEqual((30, 'col3'), result)

        # Test the dict cursor
        self.mysqldb.close()
        self.mysqldb = MySQLDB(host='localhost',
                               user='unittest',
                               db='unittest',
                               cursor_class=pymysql_utils.Cursors.DICT)
        
        for result in self.mysqldb.query('SELECT col1,col2 FROM unittest'):
          
            self.assertIsInstance(result, dict)
            
            if result['col1'] == 10:
                self.assertEqual(result['col2'], 'col1')
            elif result['col1'] == 20:
                self.assertEqual(result['col2'], 'col2')
            elif result['col1'] == 30:
                self.assertEqual(result['col2'], 'col3')

    #-------------------------
    # Query Unparameterized 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testExecuteArbitraryQuery(self):
        self.buildSmallDb()
        self.mysqldb.execute("UPDATE unittest SET col1=120")
        for result in self.mysqldb.query('SELECT col1 FROM unittest'):
            self.assertEqual(120, result)
        
    #-------------------------
    # Query Parameterized 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testExecuteArbitraryQueryParameterized(self):
        self.buildSmallDb()
        myVal = 130
        self.mysqldb.executeParameterized("UPDATE unittest SET col1=%s", (myVal,))
        for result in self.mysqldb.query('SELECT col1 FROM unittest'):
            self.assertEqual(130, result)
        
        # Provoke an error:
        (errors,warnings) = self.mysqldb.executeParameterized("UPDATE unittest SET col10=%s", (myVal,)) #@UnusedVariable
        self.assertEqual(len(errors), 1)
        
    #-------------------------
    # Reading System Variables 
    #--------------

    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testReadSysVariable(self):
        this_host = socket.gethostname()
        mysql_hostname = self.mysqldb.query('SELECT @@hostname').next()
        self.assertEqual(mysql_hostname, this_host)

    #-------------------------
    # User-Level Variables 
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testUserVariables(self):

        pre_foo = self.mysqldb.query("SELECT @foo").next()
        self.assertEqual(pre_foo, None)
        
        self.mysqldb.execute("SET @foo = 'new value';")
        
        post_foo = self.mysqldb.query("SELECT @foo").next()
        self.assertEqual(post_foo, 'new value')
        
        self.mysqldb.execute("SET @foo = 'NULL';")

    #-------------------------
    # testDbName 
    #--------------

    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testDbName(self):
        self.assertEqual(self.mysqldb.dbName(), 'unittest')
    
            
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testWithMySQLPassword(self):
        
        try:
            # Set a password for the unittest user:
            if self.mysql_ge_7:
                self.mysqldb.execute("SET PASSWORD FOR unittest@localhost = 'foobar'")
            else:
                self.mysqldb.execute("SET PASSWORD FOR unittest@localhost = PASSWORD('foobar')")

            self.mysqldb.close()
            
            # We should be unable to log in without a pwd:
            with self.assertRaises(ValueError):
                self.mysqldb = MySQLDB(host='localhost', user='unittest', db='unittest')
                
            # Open new pymysql_db.MySQLDb instance, supplying pwd: 
            self.mysqldb = MySQLDB(host='localhost', user='unittest', passwd='foobar', db='unittest')
            # Do a test query:
            self.buildSmallDb()
            res = self.mysqldb.query("SELECT col2 FROM unittest WHERE col1 = 10;").next()
            self.assertEqual(res, 'col1')
            
            # Bulk insert is also different for pwd vs. none:
            self.testBulkInsert()
        finally:
            # Make sure the remove the pwd from user unittest,
            # so that other tests will run successfully:
            if self.mysql_ge_7:
                self.mysqldb.execute("SET PASSWORD FOR unittest@localhost = ''")
            else:
                self.mysqldb.execute("SET PASSWORD FOR unittest@localhost = PASSWORD('')")
            
    #-------------------------
    # testResultCount 
    #--------------
            
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testResultCount(self):
        self.buildSmallDb()
        query_str = 'SELECT * FROM unittest'
        self.mysqldb.query(query_str)
        self.assertEqual(self.mysqldb.result_count(query_str), 3)
    
    
    #-------------------------
    # testInterleavedQueries
    #--------------
    
    @unittest.skipIf(not TEST_ALL, "Temporarily disabled")    
    def testInterleavedQueries(self):
        
        self.buildSmallDb()
        query_str1 = 'SELECT col2 FROM unittest ORDER BY col1'
        query_str2 = 'SELECT col2 FROM unittest WHERE col1 = 20 or col1 = 30 ORDER BY col1' 
        res_it1 = self.mysqldb.query(query_str1)
        res_it2 = self.mysqldb.query(query_str2)
        
        self.assertEqual(res_it1.result_count(), 3)
        self.assertEqual(res_it2.result_count(), 2)
        self.assertEqual(self.mysqldb.result_count(query_str1), 3)
        self.assertEqual(self.mysqldb.result_count(query_str2), 2)
        
        self.assertEqual(res_it1.next(), 'col1')
        self.assertEqual(res_it2.next(), 'col2')
        
        self.assertEqual(res_it1.result_count(), 3)
        self.assertEqual(res_it2.result_count(), 2)
        self.assertEqual(self.mysqldb.result_count(query_str1), 3)
        self.assertEqual(self.mysqldb.result_count(query_str2), 2)
        
        self.assertEqual(res_it1.next(), 'col2')
        self.assertEqual(res_it2.next(), 'col3')
        
        self.assertEqual(res_it1.next(), 'col3')
        with self.assertRaises(StopIteration): 
            res_it2.next()
        
        with self.assertRaises(ValueError): 
            res_it2.result_count()
            
        with self.assertRaises(ValueError): 
            self.mysqldb.result_count(query_str2)
            
        
                          
    # ----------------------- UTILITIES -------------------------
    def buildSmallDb(self):
        '''
        Creates a two-col, three-row table in database
        unittest. The table is called 'unittest'.
        Returns number of rows created.
        
        ====      ======
        col1       col2
        ====      ======
         10       'col1'
         20       'col2'
         30       'col3'
        ====      ======
        
        '''
        cur = self.mysqldb.connection.cursor()
        with no_warn_no_table():
            cur.execute('DROP TABLE IF EXISTS unittest')
        cur.execute('CREATE TABLE unittest (col1 INT, col2 TEXT)')
        cur.execute("INSERT INTO unittest VALUES (10, 'col1')")
        cur.execute("INSERT INTO unittest VALUES (20, 'col2')")
        cur.execute("INSERT INTO unittest VALUES (30, 'col3')")
        self.mysqldb.connection.commit()
        cur.close()
        return 3
    
    @classmethod  
    def get_mysql_version(cls):
        '''
        Return a tuple: (major, minor). 
        Example, for MySQL 5.7.15, return (5,7).
        Return (None,None) if version number not found.

        '''
        
        # Where is mysql client program?
        mysql_path = which('mysql')
      
        # Get version string, which looks like this:
        #   'Distrib 5.7.15, for osx10.11 (x86_64) using  EditLine wrapper\n'
        version_str = subprocess.check_output([mysql_path, '--version'])
        
        # Isolate the major and minor version numbers (e.g. '5', and '7')
        pat = re.compile(r'Distrib ([0-9]*)[.]([0-9]*)[.]')
        match_obj = pat.search(version_str)
        if match_obj is None:
            return (None,None)
        (major, minor) = match_obj.groups()
        return (int(major), int(minor))
      
        
#         self.mysqldb.dropTable('unittest')
#         self.mysqldb.createTable('unittest', schema)
#         colNames = ['col1', 'col2']
#         colValues = [(10, 'col1'), (20, 'col2'), (30, 'col3')]
#         warnings = self.mysqldb.bulkInsert('unittest', colNames, colValues)
#         self.assertIsNone(warnings)
#         return 3

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testQuery']
    unittest.main()



'''
Created on Sep 24, 2013

@author: paepcke


Modifications:
  - Dec 30, 2013: Added closing of connection to close() method
  - Mar 26, 2017: Major overhaul; fixed bulk insert.

For usage details, see `Github README <https://github.com/paepcke/pymysql_utils>`_.
This module is designed for MySQL 5.6 and 5.7. 
 
'''

from contextlib import contextmanager
import csv
import os
import re
import socket
import subprocess
import tempfile
from warnings import filterwarnings, resetwarnings

import MySQLdb
from MySQLdb import Warning as db_warning
from MySQLdb.cursors import DictCursor as DictCursor
from MySQLdb.cursors import SSCursor as SSCursor
from MySQLdb.cursors import SSDictCursor as SSDictCursor

from _mysql_exceptions import ProgrammingError, OperationalError


class DupKeyAction:
    PREVENT = 0
    IGNORE  = 1
    REPLACE = 2

# Cursor classes as per: 
#    http://mysql-python.sourceforge.net/MySQLdb-1.2.2/public/MySQLdb.cursors.BaseCursor-class.html
# Used to pass to query() method if desired:   

class Cursors:
    BASIC   			  = None
    DICT     			  = DictCursor
    SS_CURSOR       = SSCursor
    SS_DICT_CURSOR  = SSDictCursor
                      
class MySQLDB(object):
    '''
    Shallow interface to MySQL databases. Some niceties nonetheless.
      - The query() method is an iterator. So::
            for result in db.query('SELECT * FROM foo'):
               print result
      - Table manipulations (create/drop/...)
      - Insertion/update
    '''

    NON_ASCII_CHARS_PATTERN = re.compile(r'[^\x00-\x7F]+')

    # ----------------------- Top-Level Housekeeping -------------------------

    #-------------------------
    # Constructor
    #--------------

    def __init__(self, 
                 host='127.0.0.1', 
                 port=3306, 
                 user='root', 
                 passwd='', 
                 db='mysql',
                 cursor_class=None):
        '''
        Creates connection to the underlying MySQL database.
        This connection is maintained until method close()
        is called. The optional cursor_class controls the
        format in which query results are returned. The 
        Values must be one of Cursors.DICT, Cursors.SS_CURSOR,
        etc. See definition of class Cursors above. 
        
        :param host: MySQL host
        :type host: string
        :param port: MySQL host's port
        :type port: int
        :param user: user to log in as
        :type user: string
        :param passwd: password to use for given user
        :type passwd: string
        :param db: database to connect to within server
        :type db: string
        :param cursor_class: choice of how rows are returned
        :type cursor_class: Cursors
    
        '''
        
        # If all arguments are set to None, we are unittesting:
        if all(arg is None for arg in (host,port,user,passwd,db)):
            return
        
        if cursor_class is not None:
            # Ensure we caller passed a valid cursor class:
            if cursor_class not in [DictCursor, SSCursor, SSDictCursor]:
                raise ValueError("Non-existing cursor class '%s'" % str(cursor_class))
        
        self.user = user
        self.pwd  = passwd
        self.db   = db
        self.name = db
        # Will hold querySt->cursor for two purposes:
        # Ability to retrieve number of results in SELECT,
        # Ensure that all cursors are closed (see query()):
        self.cursors = {}
        self.most_recent_query = None
        
        # Find location of mysql client program:
        try:
            # The following doesn't work when running in Eclipse (gives usage for 'which'),
            # because it is running in the system env, not the bash env which
            # its usually enriched PATH var.
            self.mysql_loc = subprocess.check_output(['which', 'mysql']).strip()
        except subprocess.CalledProcessError as e:
            # Last resort: try /usr/local/bin/mysql:
            if os.path.exists('/usr/local/bin/mysql'):
                self.mysql_loc = '/usr/local/bin/mysql'
            else:
                raise RuntimeError("MySQL client not found on this machine (%s)" % socket.gethostname())

        try:
            if cursor_class is None:
                self.connection = MySQLdb.connect(host=host,
                                                  port=port, 
                                                  user=user, 
                                                  passwd=passwd, 
                                                  db=db,
                                                  charset='utf8')
            else:
                self.connection = MySQLdb.connect(host=host, 
                                                  port=port, 
                                                  user=user, 
                                                  passwd=passwd, 
                                                  db=db,
                                                  charset='utf8',
                                                  cursorclass=cursor_class)
        
        except OperationalError as e:
            pwd = '...............' if len(passwd) > 0 else '<no password>'
            raise ValueError('Cannot reach MySQL server with host:%s, port:%s, user:%s, pwd:%s, db:%s (%s)' %
                             (host, port, user, pwd, db, `e`))

    #-------------------------
    # dbName 
    #--------------
        
    def dbName(self):
        '''
        Return name of database to which this MySQLDB is connected.
        
        :return: name of database within MySQL server to which MySQLDB instance is connected.
        :rtype: String
        '''
        return self.db
    
    #-------------------------
    # close 
    #--------------

    def close(self):
        '''
        Close all cursors that are currently still open.
        '''
        for cursor in self.cursors.values():
            try:
                cursor.close()
            except:
                pass
        try:
            self.connection.close()
        except:
            pass

    # ----------------------- Table Management -------------------------

    #-------------------------
    # createTable
    #--------------
    
    def createTable(self, tableName, schema, temporary=False):
        '''
        Create new table, given its name, and schema.
        The schema is a dict mappingt column names to 
        column types. Example: {'col1' : 'INT', 'col2' : 'TEXT'}

        :param tableName: name of new table
        :type tableName: String
        :param schema: dictionary mapping column names to column types
        :type schema: Dict<String,String>
        :raise: ValueError: if table to truncate does not exist or permissions issues.        
        '''
        colSpec = ''
        for colName, colVal in schema.items():
            colSpec += str(colName) + ' ' + str(colVal) + ','
        cmd = 'CREATE %s TABLE IF NOT EXISTS %s (%s) ' % (
            'TEMPORARY' if temporary else '',
            tableName, 
            colSpec[:-1]
            )
        cursor = self.connection.cursor()
        try:
            cursor.execute(cmd)
            self.connection.commit()
        finally:
            cursor.close()

    #-------------------------
    # dropTable
    #--------------

    def dropTable(self, tableName):
        '''
        Delete table safely. No errors

        :param tableName: name of table
        :type tableName: String
        :raise: ValueError: if table to drop does not exist or permissions issues.        
        '''
        cursor = self.connection.cursor()
        try:
            # Suppress warning about table not existing:
            with no_warn_no_table():
                cursor.execute('DROP TABLE IF EXISTS %s' % tableName)
            self.connection.commit()
        except OperationalError as e:
            raise ValueError("In pymysql_utils dropTable(): %s" % `e`)
        except ProgrammingError as e:
            raise ValueError("In pymysql_utils dropTable(): %s" % `e`)
        finally:
            cursor.close()

    #-------------------------
    # truncateTable 
    #--------------

    def truncateTable(self, tableName):
        '''
        Delete all table rows. No errors

        :param tableName: name of table
        :type tableName: String
        :raise: ValueError: if table to truncate does not exist or permissions issues.
        '''
        cursor = self.connection.cursor()
        try:
            try:
                cursor.execute('TRUNCATE TABLE %s' % tableName)
            except OperationalError as e:
                raise ValueError("In pymysql_utils truncateTable(): %s." % `e`)
            except ProgrammingError as e:
                raise ValueError("In pymysql_utils truncateTable(): %s." % `e`)
            self.connection.commit()
        finally:
            cursor.close()

    # ----------------------- Insertion and Updates -------------------------

    #-------------------------
    # insert 
    #--------------

    def insert(self, tblName, colnameValueDict):
        '''
        Given a dictionary mapping column names to column values,
        insert the data into a specified table

        :param tblName: name of table to insert into
        :type tblName: String
        :param colnameValueDict: mapping of column name to column value
        :type colnameValueDict: Dict<String,Any>
        :return None if all ok, else tuple (errorList, warningsList)
        '''
        colNames, colValues = zip(*colnameValueDict.items())
        cursor = self.connection.cursor()
        try:
            wellTypedColValues = self._ensureSQLTyping(colValues)
            cmd = 'INSERT INTO %s (%s) VALUES (%s)' % (str(tblName), ','.join(colNames), wellTypedColValues)
            with no_db_warnings():
                try:
                    cursor.execute(cmd)
                except Exception:
                    # The following show_warnings() will
                    # reveal the error:
                    pass
            mysql_warnings = self.connection.show_warnings()
            if len(mysql_warnings) > 0:
                warnings   = [warning_tuple for warning_tuple in mysql_warnings if warning_tuple[0] == 'Warning']
                errors     = [error_tuple for error_tuple in mysql_warnings if error_tuple[0] == 'Error']
        finally:
            self.connection.commit()
            cursor.close()
            return (None,None) if len(mysql_warnings) == 0 else (errors, warnings)            
    
    #-------------------------
    # bulkInsert 
    #--------------
    
    def bulkInsert(self, tblName, colNameTuple, valueTupleArray, onDupKey=DupKeyAction.PREVENT):
        '''
        Inserts large number of rows into given table. The
        rows must be provided as an array of tuples. Caller's
        choice whether duplicate keys should cause a warning,
        omitting the incoming row, omit the incoming row without
        a warning, or replace the existing row.
        
        Returns None if all went well, else a list of error
        tuples.
        
        Strategy: 
        
        write the values to a temp file, then generate a 
        ``LOAD LOCAL INFILE...`` MySQL command and execute it in MySQL. 
        Returns `None` if no errors/warnings, else returns the tuple
        of tuples with the warnings.
                
        :param tblName: table into which to insert
        :type  tblName: string
        :param colNameTuple: tuple containing column names in proper order, i.e. 
               corresponding to valueTupleArray orders.
        :type  colNameTuple: (str[,str[...]])
        :param valueTupleArray: array of n-tuples, which hold the values. Order of
               values must correspond to order of column names in colNameTuple.
        :type  valueTupleArray: `[(<MySQLVal> [,<MySQLval>,...]])`
        :param onDupKey: determines action when incoming row would duplicate an existing row's
        	   unique key. If set to DupKeyAction.IGNORE, then the incoming tuple is
        	   skipped. If set to DupKeyAction.REPLACE, then the incoming tuple replaces
        	   the existing tuple. By default, each attempt to insert a duplicate key
        	   generates a warning.
        :type  onDupKey: DupKeyAction
        :return: None is no warnings occurred, else 
            a tuple with (errorList, warningsList).
        		Warning reflects MySQL's output of "show warnings;"
        		Example::
        		((u'Warning', 1062L, u"Duplicate entry '10' for key 'PRIMARY'"),)
        				    
        :rtype: {None | ({ (str) | None},{ (str) | None})}
        :raise: ValueError if bad parameter.
        
        '''
        tmpCSVFile = tempfile.NamedTemporaryFile(dir='/tmp',prefix='bulkInsert',suffix='.csv')
        # Allow MySQL to read this tmp file:
        os.chmod(tmpCSVFile.name, 0644)
        self.csvWriter = csv.writer(tmpCSVFile, 
                                    dialect='excel-tab', 
                                    lineterminator='\n', 
                                    delimiter=',', 
                                    quotechar='"', 
                                    quoting=csv.QUOTE_MINIMAL)
        # Can't use csvWriter.writerows() b/c some rows have 
        # weird chars: self.csvWriter.writerows(valueTupleArray)        
        for row in valueTupleArray:
            # Convert each element in row to a string,
            # including mixed-in Unicode Strings:
            self.csvWriter.writerow([rowElement for rowElement in self._stringifyList(row)])
        tmpCSVFile.flush()
        
        # Create the MySQL column name list needed in the LOAD INFILE below.
        # We need '(colName1,colName2,...)':
        if len(colNameTuple) == 0:
            colSpec = '()'
        else:
            colSpec = '(' + colNameTuple[0]
            for colName in colNameTuple[1:]:
                colSpec += ',' + colName
            colSpec += ')'

        # For warnings from MySQL:
        mysql_warnings = ''
        try:
            if onDupKey == DupKeyAction.PREVENT:
                dupAction = '' # trigger the MySQL default
            elif onDupKey == DupKeyAction.IGNORE:
                dupAction = 'IGNORE'
            elif onDupKey == DupKeyAction.REPLACE:
                dupAction = 'REPLACE'
            else:
                raise ValueError("Parameter onDupKey to bulkInsert method must be of type DupKeyAction; is %s" % str(onDupKey))
            
            # Remove quotes from the values inside the colNameTuple's:
            mySQLCmd = ("LOAD DATA LOCAL INFILE '%s' %s INTO TABLE %s FIELDS TERMINATED BY ',' " +\
                        "OPTIONALLY ENCLOSED BY '\"' ESCAPED BY '\\\\' LINES TERMINATED BY '\\n' %s"
                        ) %  (tmpCSVFile.name, dupAction, tblName, colSpec)
            cursor = self.connection.cursor()
            #with no_warn_dup_key():
            with no_db_warnings():
                try:
                    cursor.execute(mySQLCmd)
                except Exception:
                    # The following show_warnings() will
                    # reveal the error:
                    pass
            mysql_warnings = self.connection.show_warnings()
            if len(mysql_warnings) > 0:
                warnings   = [warning_tuple for warning_tuple in mysql_warnings if warning_tuple[0] == 'Warning']
                errors     = [error_tuple for error_tuple in mysql_warnings if error_tuple[0] == 'Error']
                if len(warnings) == 0:
                    warnings = None
                if len(errors) == 0:
                    errors = None
        finally:
            tmpCSVFile.close()
            self.execute('commit;')
            return (None,None) if len(mysql_warnings) == 0 else (errors, warnings)            
    
    #-------------------------
    # update 
    #--------------
    
    def update(self, tblName, colName, newVal, fromCondition=None):
        '''
        Update one column with a new value.

        :param tblName: name of table in which update is to occur
        :type tblName: String
        :param colName: column whose value is to be changed
        :type colName: String
        :param newVal: value acceptable to MySQL for the given column 
        :type newVal: type acceptable to MySQL for the given column 
        :param fromCondition: optionally condition that selects which rows to update.
                      if None, the named column in all rows are updated to
                      the given value. Syntax must conform to what may be in
                      a MySQL FROM clause (don't include the 'FROM' keyword)
        :type fromCondition: String
        :return None if all ok, else tuple (errorList, warningsList)
        '''
        cursor = self.connection.cursor()
        try:
            if fromCondition is None:
                cmd = "UPDATE %s SET %s = '%s';" % (tblName,colName,newVal)
            else:
                cmd = "UPDATE %s SET %s = '%s' WHERE %s;" % (tblName,colName,newVal,fromCondition)
            with no_db_warnings():
                try:
                    cursor.execute(cmd)
                except Exception:
                    # The following show_warnings() will
                    # reveal the error:
                    pass
            mysql_warnings = self.connection.show_warnings()
            if len(mysql_warnings) > 0:
                warnings   = [warning_tuple for warning_tuple in mysql_warnings if warning_tuple[0] == 'Warning']
                errors     = [error_tuple for error_tuple in mysql_warnings if error_tuple[0] == 'Error']
                if len(warnings) == 0:
                    warnings = None
                if len(errors) == 0:
                    errors = None

        finally:
            self.connection.commit()            
            cursor.close()
            return (None,None) if len(mysql_warnings) == 0 else (errors, warnings)            

    # ----------------------- Queries -------------------------
        
    #-------------------------
    # query 
    #--------------
    
    def query(self, queryStr):
        '''
        Query iterator. Given a query, return one result for each
        subsequent call. When all results have been retrieved,
        the iterator returns None.
        
        IMPORTANT: what is returned is an *iterator*. So you
        need to call next() to get the first result, even if
        there is only one.

        :param queryStr: the query to submit to MySQL
        :type queryStr: String
        :return: iterator of query results
        :rtype: iterator of tuple
        '''

        queryStr = queryStr.encode('UTF-8')
        
        # For if caller never exhausts the results by repeated calls,
        # and to find cursor by query to get (e.g.) num results:
        cursor = self.connection.cursor()
        
        # Ability to find an active cursor by query 
        # via result_count(queryString)...
        self.cursors[queryStr] = cursor
        
        #... or find it via  result_count()...
        self.most_recent_query = queryStr

        try:        
            cursor.execute(queryStr)
        except ProgrammingError as e:
            raise ValueError(`e`)
        return QueryResult(cursor, queryStr, self)
        
    #-------------------------
    # result_count 
    #--------------

    def result_count(self, queryStr=None):
        '''
        Given a query string, after that string served
        as query in a call to query(), and before the result iterator
        has been exhausted: return number of SELECT results.
        
        If queryStr is left to None, the value of 
        self.most_recent_query is used.
        
        :param queryStr: query that was used in a prior call to query()
        :type queryStr: {None | string}
        :return: number of records in SELECT result.
        :rtype: int
        :raise: ValueError if no prior query is still active.
        '''
        
        try:
            if queryStr is None:
                if self.most_recent_query is not None:
                    return self.cursors[self.most_recent_query].rowcount
                else:
                    raise(ValueError("No previous query available."))
            else:
                return self.cursors[queryStr].rowcount
        except KeyError:
            raise ValueError("Query '%s' is no longer active." % queryStr)

    
    #-------------------------
    # execute 
    #--------------
            
    def execute(self,query, doCommit=True):                                                                                   
        '''
        Execute an arbitrary query, including
        MySQL directives. Return value is undefined.
        For SELECT queries or others that return
        results from MySQL, use the query() method. 
        
        Note: For some funky directives you get the MySQL error:
        
            Commands out of sync; you can't run this command now 
            
        For those, turn doCommit to False;
        
        :param query: query or directive
        :type query: String
        :return: None if all went well, else a tuple:
            (listOrErrors, listOfWarnings)
        '''
        
        cursor=self.connection.cursor()                                                                        
        try:
            with no_db_warnings():
                try:                                                                                                   
                    cursor.execute(query)
                except Exception:
                    # The following show_warnings() will
                    # reveal the error:
                    pass
            mysql_warnings = self.connection.show_warnings()
            if len(mysql_warnings) > 0:
                warnings   = [warning_tuple for warning_tuple in mysql_warnings if warning_tuple[0] == 'Warning']
                errors     = [error_tuple for error_tuple in mysql_warnings if error_tuple[0] == 'Error']
                if len(warnings) == 0:
                    warnings = None
                if len(errors) == 0:
                    errors = None
        finally:                                                                                               
            if doCommit:                                                                              
                self.connection.commit()                                                                           
            cursor.close()                                                                                     
            return (None,None) if len(mysql_warnings) == 0 else (errors, warnings)

    #-------------------------
    # executeParameterized
    #--------------
    
    def executeParameterized(self,query,params):                                                                      
        '''
        Executes arbitrary query that is parameterized
        as in the Python string format statement. Ex:
        executeParameterized('SELECT %s FROM myTable', ('col1', 'col3'))
        
        Return value is undefined. For SELECT queries or others that return
        results from MySQL, use the query() method.        

        Note: params must be a tuple. Example: to update a column:: 
        mysqldb.executeParameterized("UPDATE myTable SET col1=%s", (myVal,))
        the comma after 'myVal' is mandatory; it indicates that the 
        expression is a tuple.

        :param   query: query with parameter placeholder
        :type    query: string
        :param   params: tuple of actuals for the parameters.
        :type    params: (<any>)
        :return: None if all ok, else tuple: (errorList, warningsList)
                
        '''
        cursor=self.connection.cursor()                                                                        
        try:                                                                                                   
            with no_db_warnings():
                try:
                    cursor.execute(query,params)
                except Exception:
                    # The following show_warnings() will
                    # reveal the error:
                    pass
            mysql_warnings = self.connection.show_warnings()
            if len(mysql_warnings) > 0:
                warnings   = [warning_tuple for warning_tuple in mysql_warnings if warning_tuple[0] == 'Warning']
                errors     = [error_tuple for error_tuple in mysql_warnings if error_tuple[0] == 'Error']
                if len(warnings) == 0:
                    warnings = None
                if len(errors) == 0:
                    errors = None
        finally:
            self.connection.commit()                                                                                                                                                                          
            cursor.close()
            return (None,None) if len(mysql_warnings) == 0 else (errors, warnings)

    # ----------------------- Utilities -------------------------                    


    #-------------------------
    # query_exhausted 
    #--------------

    def query_exhausted(self, cursor):

        # Delete the cursor's entry in the 
        # cursors { queryStr --> cursor } dict.
        # Since there will be few entries, the
        # following dictionary scan isn't a problem:
    
        for query_str, the_cursor in self.cursors.items():
            if the_cursor == cursor:
                del self.cursors[query_str]
        cursor.close()
    
    #-------------------------
    # _ensureSQLTyping
    #--------------
    
    def _ensureSQLTyping(self, colVals):
        '''
        Given a list of items, return a string that preserves
        MySQL typing. Example: (10, 'My Poem') ---> '10, "My Poem"'
        Note that ','.join(map(str,myList)) won't work:
        (10, 'My Poem') ---> '10, My Poem'

        :param colVals: list of column values destined for a MySQL table
        :type colVals: <any>
        :return: string of string-separated, properly typed column values
        :rtype: string
        '''

        resList = []
        for el in colVals:
            if isinstance(el, basestring):
                try:
                    # If value is not already UTF-8, encode it:
                    cleanStr = unicode(el, 'UTF-8', 'replace')
                except TypeError:
                    # Value was already in Unicode, so all is well:
                    cleanStr = el
                resList.append('"%s"' % cleanStr)
            elif el is None:
                resList.append('null')
            elif isinstance(el, (list, dict, set)):
                resList.append('"%s"' % str(el))
            else: # e.g. numbers
                resList.append(el)
        try:
            return ','.join(map(unicode,resList))
        except UnicodeEncodeError as e:
            print('Unicode related error: %s' % `e`)
            
    #-------------------------
    # _stringifyList
    #--------------
    
    def _stringifyList(self, iterable):
        '''
        Goes through the iterable. For each element, tries
        to turn into a string, part of which attempts encoding
        with the 'ascii' codec. Then encountering a unicode
        char, that char is UTF-8 encoded.
        
        Acts as an iterator! Use like:
        for element in _stringifyList(someList):
            print(element)

        :param iterable: mixture of items of any type, including Unicode strings.
        :type iterable: [<any>]
        '''
        for element in iterable:
            try:
                yield(str(element))
            except UnicodeEncodeError:
                yield element.encode('UTF-8','ignore')

# ----------------------- Class QueryResult -------------------------

class QueryResult(object):
    '''
    Iterator for query results. Given an instance
    of this class I, the following works as expected:

        for res in I:
            print(res)
            
    Instances of this class are returned by MySQLDB's 
    query() method. Use next() and nextall() to get
    one result at a time, or all at once.
    '''
  
    def __init__(self, cursor, query_str, cursor_owner_obj):
        self.mysql_cursor = cursor
        self.cursor_owner = cursor_owner_obj
        self.the_query_str    = query_str
        self.exhausted    = False
      
    def __iter__(self):
        return self

    def next(self):
        '''
        Return the next result in a query, 
        or raise StopIteration if no results
        remain.
        
        Note: Returns the zeroeth element
        of the result. So:
        `('foo',)  --> 'foo'` 
        and 
        `({'col1' : 10},) ==> {'col1' : 10}`

        This convention is in contrast to what
        cursor.fetconeh() does.
        
        '''
  
        res = self.mysql_cursor.fetchone()
        if res is None:
            self.cursor_owner.query_exhausted(self.mysql_cursor)
            self.exhausted = True
            raise StopIteration()            
        else:
            if len(res) == 1:
                return res[0]
            else:
                return res
          
    def nextall(self):
        '''
        Returns the remaining query results as a 
        tuple of tuples.
        
        :return: all remaining tuples inside a wrapper tuple, or empty tuple
                 if no results remain.
        :rtype: ((str))
        
        '''
        all_remaining = self.mysql_cursor.fetchall()
        # We exhausted the query, so clean up:
        self.cursor_owner.query_exhausted(self.mysql_cursor)
        self.exhausted = True
        return all_remaining

    def query_str(self):
        '''
        Provide query that led to this result.
        '''
        return self.the_query_str

    def result_count(self):
        '''
        Return the number of results in this result object.
        '''
        if self.exhausted:
            raise ValueError("Query '%s' is no longer active." % self.query_str())
        return self.mysql_cursor.rowcount
      
      
# ----------------------- Context Managers -------------------------    

# Ability to write:
#    with no_warn_no_table():
#       ... DROP TABLE IF NOT EXISTS ...
# without annoying Python-level warnings that the
# table did not exist:

@contextmanager
def no_warn_no_table():
    filterwarnings('ignore', message="Unknown table", category=db_warning)
    yield
    resetwarnings()

@contextmanager
def no_warn_dup_key():
    filterwarnings('ignore', message="Duplicate entry", category=db_warning)
    yield
    resetwarnings()

@contextmanager
def no_db_warnings():
    filterwarnings('ignore', category=db_warning)
    yield
    resetwarnings()
